import base64
from hashlib import sha256 as SHA256
import hmac as HMAC
import json
import logging
import os
import pathlib
import struct
from typing import Dict, Tuple, Union

from pbkdf2 import PBKDF2
from pyaes import AESModeOfOperationCBC, Encrypter, Decrypter


logger = logging.getLogger('audible.aescipher')


class AESCipher:
    """Encrypt/Decrypt data using password to generate key. The encryption
    algorithm used is symmetric AES in cipher-block chaining (CBC) mode.
    
    ``key_size`` may be 16, 24 or 32 (default).
    The key is derived via the PBKDF2 key derivation function (KDF) from the
    password and a random salt of 16 bytes (the AES block size) minus the
    length of the salt header (see below).
    The hash function used by PBKDF2 is SHA256 per default. You can pass a
    different hash function module via the ``hashmod`` argument. The module
    must adhere to the Python API for Cryptographic Hash Functions (PEP 247).
    PBKDF2 uses a number of iterations of the hash function to derive the key,
    which can be set via the ``kdf_iterations` keyword argumeent. The default
    number is 1000 and the maximum 65535.
    The header and the salt are written to the first block of the encrypted
    output (bytes mode) or written as key/value pairs (dict mode). The header
    consist of the number of KDF iterations encoded as a big-endian word bytes
    wrapped by ``salt_marker`` on both sides. With the default value of
    ``salt_marker = b'$'``, the header size is thus 4 and the salt 12 bytes.
    The salt marker must be a byte string of 1-6 bytes length.
    The last block of the encrypted output is padded with up to 16 bytes, all
    having the value of the length of the padding.
    All values in dict mode are written as base64 encoded string.
    """

    def __init__(self, password: str, *, key_size: int = 32,
                 salt_marker: bytes = b"$", kdf_iterations: int = 1000,
                 hashmod=SHA256, mac=HMAC) -> None:

        self.password = password
        self.key_size = key_size
        self.hashmod = hashmod
        self.mac = mac
        self.bs = 16
        
        if not 1 <= len(salt_marker) <= 6:
            raise ValueError('The salt_marker must be one to six bytes long.')
        elif not isinstance(salt_marker, bytes):
            raise TypeError('salt_marker must be a bytes instance.')
        else:
            self.salt_marker = salt_marker

        if kdf_iterations >= 65536:
            raise ValueError('kdf_iterations must be <= 65535.')
        else:
            self.kdf_iterations = kdf_iterations

    def _encrypt(self, data: str) -> Tuple[bytes, bytes, bytes]:
        bs = self.bs
        header = (self.salt_marker
                  + struct.pack('>H', self.kdf_iterations)
                  + self.salt_marker)
        salt = os.urandom(bs - len(header))
        kdf = PBKDF2(self.password, salt, min(self.kdf_iterations, 65535),
                     self.hashmod, self.mac)
        key = kdf.read(self.key_size)
        iv = os.urandom(bs)
        encrypter = Encrypter(AESModeOfOperationCBC(key, iv))

        encrypted = encrypter.feed(data)
        encrypted += encrypter.feed()
    
        return (header+salt), iv, encrypted

    def _decrypt(self, salt: bytes, iv: bytes, encrypted_data: bytes) -> str:
        mlen = len(self.salt_marker)
        hlen = mlen * 2 + 2
    
        if salt[:mlen] == self.salt_marker and salt[mlen + 2:hlen] == self.salt_marker:
            kdf_iterations = struct.unpack('>H', salt[mlen:mlen + 2])[0]
            salt = salt[hlen:]
        else:
            kdf_iterations = self.kdf_iterations
    
        if kdf_iterations >= 65536:
            raise ValueError('kdf_iterations must be <= 65535.')

        kdf = PBKDF2(self.password, salt, self.kdf_iterations,
                     self.hashmod, self.mac)
        key = kdf.read(self.key_size)
        decrypter = Decrypter(AESModeOfOperationCBC(key, iv))

        decrypted = decrypter.feed(encrypted_data)
        decrypted += decrypter.feed()

        return decrypted.decode("utf-8")

    def to_dict(self, data: str) -> Dict[str, str]:
        salt, iv, encrypted_data = self._encrypt(data)

        return {
            "salt": base64.b64encode(salt).decode("utf-8"),
            "iv": base64.b64encode(iv).decode("utf-8"),
            "ciphertext": base64.b64encode(encrypted_data).decode("utf-8"),
            "info": 'base64-encoded AES-CBC-256 of JSON object'
        }

    def from_dict(self, data: dict) -> str:
        salt = base64.b64decode(data["salt"])
        iv = base64.b64decode(data["iv"])
        encrypted_data = base64.b64decode(data["ciphertext"])

        return self._decrypt(salt, iv, encrypted_data)

    def to_bytes(self, data: str) -> bytes:
        salt, iv, encrypted_data = self._encrypt(data)

        return salt + iv + encrypted_data

    def from_bytes(self, data: bytes) -> str:
        bs = self.bs
        salt = data[:bs]
        iv = data[bs:2*bs]
        encrypted_data = data[2*bs:]

        return self._decrypt(salt, iv, encrypted_data)

    def to_file(self, data: str, filename: Union[pathlib.Path, pathlib.WindowsPath],
                encryption="json", indent=4):
        if encryption == "json":
            encrypted_dict = self.to_dict(data)
            data_json = json.dumps(encrypted_dict, indent=indent)
            filename.write_text(data_json)

        elif encryption == "bytes":
            encrypted_data = self.to_bytes(data)
            filename.write_bytes(encrypted_data)

        else:
            raise ValueError("encryption must be \"json\" or \"bytes\"..")

    def from_file(self, filename: Union[pathlib.Path, pathlib.WindowsPath],
                  encryption="json"):
        if encryption == "json":
            encrypted_json = filename.read_text()
            encrypted_dict = json.loads(encrypted_json)

            return self.from_dict(encrypted_dict)

        elif encryption == "bytes":
            encrypted_data = filename.read_bytes()
    
            return self.from_bytes(encrypted_data)

        else:
            raise ValueError("encryption must be \"json\" or \"bytes\".")


def detect_file_encryption(filename: Union[pathlib.Path, pathlib.WindowsPath]):
    file = filename.read_bytes()
    try:
        file = json.loads(file)
        if "adp_token" in file:
            return False
        elif "ciphertext" in file:
            return "json"
    except UnicodeDecodeError:
        return "bytes"


def remove_file_encryption(source, target, password, **kwargs):
    encryption = detect_file_encryption(source)
    if encryption:
        source_file = pathlib.Path(source)
        crypter = AESCipher(password, **kwargs)
        decrypted = crypter.from_file(source_file, encryption=encryption)
        pathlib.Path(target).write_text(decrypted)
    else:
        raise Exception("file is not encrypted")



# -*- coding: utf-8 -*-

from .client import AudibleAPI
from .auth import LoginAuthenticator, FileAuthenticator
from .localization import Locale, autodetect_locale
from ._logging import set_file_logger, set_console_logger
from ._version import *


__all__ = [
    "__version__", "AudibleAPI", "LoginAuthenticator", "FileAuthenticator",
    "Locale", "autodetect_locale", "set_file_logger", "set_console_logger",
]



import base64
import binascii
from datetime import datetime
import io
import json
import math
from typing import Any, Dict, List, Union
from urllib.parse import urlencode, parse_qs

from bs4 import BeautifulSoup
from PIL import Image
import requests


USER_AGENT = ("Mozilla/5.0 (iPhone; CPU iPhone OS 12_3_1 like Mac OS X) "
              "AppleWebKit/605.1.15 (KHTML, like Gecko) Mobile/15E148")


def default_captcha_callback(captcha_url: str) -> str:
    """Helper function for handling captcha."""
    # on error print captcha url instead of display captcha
    try:
        captcha = requests.get(captcha_url).content
        f = io.BytesIO(captcha)
        img = Image.open(f)
        img.show()
    except:
        print(captcha_url)

    guess = input("Answer for CAPTCHA: ")
    return str(guess).strip().lower()


def default_otp_callback() -> str:
    """Helper function for handling 2-factor authentication."""
    guess = input("OTP Code: ")
    return str(guess).strip().lower()


def get_soup(resp):
    return BeautifulSoup(resp.text, "html.parser")


def get_inputs_from_soup(soup) -> Dict[str, str]:
    inputs = {}
    for node in soup.select("input[type=hidden]"):
        if node["name"] and node["value"]:
            inputs[node["name"]] = node["value"]
    return inputs


def build_oauth_url(countryCode: str, domain: str, marketPlaceId: str) -> str:
    oauth_params = {
        "openid.oa2.response_type": "token",
        "openid.return_to": f"https://www.amazon.{domain}/ap/maplanding",
        "openid.assoc_handle": f"amzn_audible_ios_{countryCode}",
        "openid.identity": ("http://specs.openid.net/auth/2.0/"
                            "identifier_select"),
        "pageId": "amzn_audible_ios",
        "accountStatusPolicy": "P1",
        "openid.claimed_id": ("http://specs.openid.net/auth/2.0/"
                              "identifier_select"),
        "openid.mode": "checkid_setup",
        "openid.ns.oa2": "http://www.amazon.com/ap/ext/oauth/2",
        "openid.oa2.client_id": ("device:6a52316c62706d53427a57355"
                                 "05a76477a45375959566674327959465"
                                 "a6374424a53497069546d45234132435"
                                 "a4a5a474c4b324a4a564d"),
        "openid.ns.pape": "http://specs.openid.net/extensions/pape/1.0",
        "marketPlaceId": marketPlaceId,
        "openid.oa2.scope": "device_auth_access",
        "forceMobileLayout": "true",
        "openid.ns": "http://specs.openid.net/auth/2.0",
        "openid.pape.max_auth_age": "0"
    }
    
    return f"https://www.amazon.{domain}/ap/signin?{urlencode(oauth_params)}"


def check_for_captcha(soup):
    captcha = soup.find("img", alt=lambda x: x and "CAPTCHA" in x)
    return True if captcha else False


def extract_captcha_url(soup):
    captcha = soup.find("img", alt=lambda x: x and "CAPTCHA" in x)
    return captcha["src"] if captcha else None


def check_for_mfa(soup):
    mfa = soup.find("form", id=lambda x: x and "auth-mfa-form" in x)
    return True if mfa else False


def check_for_choice_mfa(soup):
    mfa_choice = soup.find("form", id="auth-select-device-form")
    return True if mfa_choice else False


def check_for_cvf(soup):
    cvf = soup.find("div", id="cvf-page-content")
    return True if cvf else False


def extract_cookies_from_session(session):
    cookies = dict()
    for cookie in session.cookies:
        cookies[cookie.name] = cookie.value.replace(r'"', r'')
    return cookies


def extract_token_from_url(url):
    parsed_url = parse_qs(url)
    return parsed_url["openid.oa2.access_token"][0]


def login(username: str, password: str, countryCode: str,
          domain: str, marketPlaceId: str, captcha_callback=None,
          otp_callback=None) -> Dict[str, Any]:

    amazon_url = f"https://www.amazon.{domain}"
    sign_in_url = amazon_url + "/ap/signin"
    cvf_url = amazon_url + "/ap/cvf/verify"

    session = requests.Session()
    session.headers.update({"User-Agent": USER_AGENT})

    while "session-token" not in session.cookies:
        session.get(amazon_url)

    oauth_url = build_oauth_url(countryCode, domain, marketPlaceId)
    oauth_resp = session.get(oauth_url)
    oauth_soup = get_soup(oauth_resp)

    login_inputs = get_inputs_from_soup(oauth_soup)
    login_inputs["email"] = username
    login_inputs["password"] = password
    metadata = meta_audible_app(USER_AGENT, amazon_url)
    login_inputs["metadata1"] = encrypt_metadata(metadata)

    login_resp = session.post(sign_in_url, data=login_inputs)
    login_soup = get_soup(login_resp)

    # check for captcha
    while check_for_captcha(login_soup):
        captcha_url = extract_captcha_url(login_soup)
        if captcha_callback:
            guess = captcha_callback(captcha_url)
        else:
            guess = default_captcha_callback(captcha_url)

        inputs = get_inputs_from_soup(login_soup)
        inputs["guess"] = guess
        inputs["use_image_captcha"] = "true"
        inputs["use_audio_captcha"] = "false"
        inputs["showPasswordChecked"] = "false"
        inputs["email"] = username
        inputs["password"] = password

        login_resp = session.post(sign_in_url, data=inputs)
        login_soup = get_soup(login_resp)

    # check for choice mfa
    # https://www.amazon.de/ap/mfa/new-otp
    while check_for_choice_mfa(login_soup):
        inputs = get_inputs_from_soup(login_soup)
        for node in login_soup.select("div[data-a-input-name=otpDeviceContext]"):
            # auth-TOTP, auth-SMS, auth-VOICE
            if "auth-TOTP" in node["class"]:
                inp_node = node.find("input")
                inputs[inp_node["name"]] = inp_node["value"]

        login_resp = session.post(amazon_url + "/ap/mfa", data=inputs)
        login_soup = get_soup(login_resp)

    # check for mfa (otp_code)
    while check_for_mfa(login_soup):
        if otp_callback:
            otp_code = otp_callback()
        else:
            otp_code = default_otp_callback()

            inputs = get_inputs_from_soup(login_soup)
        inputs["otpCode"] = otp_code
        inputs["mfaSubmit"] = "Submit"
        inputs["rememberDevice"] = "false"

        login_resp = session.post(sign_in_url, data=inputs)
        login_soup = get_soup(login_resp)

    # check for cvf
    while check_for_cvf(login_soup):
        inputs = get_inputs_from_soup(login_soup)

        login_resp = session.post(cvf_url, data=inputs)
        login_soup = get_soup(login_resp)

        inputs = get_inputs_from_soup(login_soup)
        inputs["action"] = "code"
        inputs["code"] = input("Code: ")

        login_resp = session.post(cvf_url, data=inputs)
        login_soup = get_soup(login_resp)

    if login_resp.status_code != 404:
        raise Exception("Unable to login")

    access_token = extract_token_from_url(login_resp.url)
    login_cookies = extract_cookies_from_session(session)

    session.close()

    return {
        "access_token": access_token,
        "login_cookies": login_cookies
    }


# constants used for encrypt/decrypt metadata
WRAP_CONSTANT = 2654435769
CONSTANTS = [1888420705, 2576816180, 2347232058, 874813317]


def _data_to_int_list(data: Union[str, bytes]) -> List[int]:
    data_bytes = data.encode() if isinstance(data, str) else data
    data_list_int = []
    for i in range(0, len(data_bytes), 4):
        data_list_int.append(int.from_bytes(data_bytes[i:i+4], "little"))

    return data_list_int


def _list_int_to_bytes(data: List[int]) -> bytearray:
    data_list = data
    data_bytes = bytearray()
    for i in data_list:
        data_bytes += i.to_bytes(4, 'little')

    return data_bytes


def _encrypt_data(data: List[int]) -> List[int]:
    temp2 = data
    rounds = len(temp2)
    minor_rounds = int(6 + (52 / rounds // 1))
    last = temp2[-1]
    inner_roll = 0

    for _ in range(minor_rounds):
        inner_roll += WRAP_CONSTANT
        inner_variable = inner_roll >> 2 & 3

        for i in range(rounds):
            first = temp2[(i + 1) % rounds]
            temp2[i] += ((last >> 5 ^ first << 2)
                         + (first >> 3 ^ last << 4) ^ (inner_roll ^ first)
                         + (CONSTANTS[i & 3 ^ inner_variable] ^ last))
            last = temp2[i] = temp2[i] & 0xffffffff

    return temp2


def _decrypt_data(data: List[int]) -> List[int]:
    temp2 = data
    rounds = len(temp2)
    minor_rounds = int(6 + (52 / rounds // 1))
    inner_roll = 0

    for _ in range(minor_rounds + 1):
        inner_roll += WRAP_CONSTANT
        inner_variable = inner_roll >> 2 & 3

    for _ in range(minor_rounds):
        inner_roll -= WRAP_CONSTANT
        inner_variable = inner_roll >> 2 & 3

        for i in range(rounds):
            i = rounds - i - 1

            first = temp2[(i + 1) % rounds]
            last = temp2[(i - 1) % rounds]

            temp2[i] -= ((last >> 5 ^ first << 2)
                         + (first >> 3 ^ last << 4) ^ (inner_roll ^ first)
                         + (CONSTANTS[i & 3 ^ inner_variable] ^ last))
            last = temp2[i] = temp2[i] & 0xffffffff

    return temp2


def _generate_hex_checksum(data: str) -> str:
    checksum_int = binascii.crc32(data.encode()) & 0xffffffff
    return checksum_int.to_bytes(4, byteorder='big').hex().upper()


def encrypt_metadata(metadata: str) -> str:
    """
    Encrypts metadata to be used to log in to amazon.

    """
    checksum = _generate_hex_checksum(metadata)

    object_str = f"{checksum}#{metadata}"

    object_list_int = _data_to_int_list(object_str)

    object_list_int_enc = _encrypt_data(object_list_int)

    object_bytes = _list_int_to_bytes(object_list_int_enc)

    object_base64 = base64.b64encode(object_bytes)

    return f"ECdITeCs:{object_base64.decode()}"


def decrypt_metadata(metadata: str) -> str:
    """Decrypt metadata. For testing purposes only."""
    object_base64 = metadata.lstrip("ECdITeCs:")

    object_bytes = base64.b64decode(object_base64)

    object_list_int = _data_to_int_list(object_bytes)

    object_list_int_dec = _decrypt_data(object_list_int)

    object_bytes = _list_int_to_bytes(object_list_int_dec).rstrip(b"\0")

    object_str = object_bytes.decode()

    checksum, metadata = object_str.split("#", 1)

    assert _generate_hex_checksum(metadata) == checksum

    return metadata


def now_to_unix_ms() -> int:
    return math.floor(datetime.now().timestamp()*1000)


def meta_audible_app(user_agent: str, oauth_url: str) -> str:
    """
    Returns json-formatted metadata to simulate sign-in from iOS audible app.
    """

    meta_dict = {
        "start": now_to_unix_ms(),
        "interaction": {
            "keys": 0,
            "keyPressTimeIntervals": [],
            "copies": 0,
            "cuts": 0,
            "pastes": 0,
            "clicks": 0,
            "touches": 0,
            "mouseClickPositions": [],
            "keyCycles": [],
            "mouseCycles": [],
            "touchCycles": []
        },
        "version": "3.0.0",
        "lsUbid": "X39-6721012-8795219:1549849158",
        "timeZone": -6,
        "scripts": {
            "dynamicUrls": [
                ("https://images-na.ssl-images-amazon.com/images/I/"
                 "61HHaoAEflL._RC|11-BZEJ8lnL.js,01qkmZhGmAL.js,71qOHv6nKaL."
                 "js_.js?AUIClients/AudibleiOSMobileWhiteAuthSkin#mobile"),
                ("https://images-na.ssl-images-amazon.com/images/I/"
                 "21T7I7qVEeL._RC|21T1XtqIBZL.js,21WEJWRAQlL.js,31DwnWh8lFL."
                 "js,21VKEfzET-L.js,01fHQhWQYWL.js,51TfwrUQAQL.js_.js?"
                 "AUIClients/AuthenticationPortalAssets#mobile"),
                ("https://images-na.ssl-images-amazon.com/images/I/"
                 "0173Lf6yxEL.js?AUIClients/AuthenticationPortalInlineAssets"),
                ("https://images-na.ssl-images-amazon.com/images/I/"
                 "211S6hvLW6L.js?AUIClients/CVFAssets"),
                ("https://images-na.ssl-images-amazon.com/images/G/"
                 "01/x-locale/common/login/fwcim._CB454428048_.js")
            ],
            "inlineHashes": [
                -1746719145,
                1334687281,
                -314038750,
                1184642547,
                -137736901,
                318224283,
                585973559,
                1103694443,
                11288800,
                -1611905557,
                1800521327,
                -1171760960,
                -898892073
            ],
            "elapsed": 52,
            "dynamicUrlCount": 5,
            "inlineHashesCount": 13
        },
        "plugins": "unknown||320-568-548-32-*-*-*",
        "dupedPlugins": "unknown||320-568-548-32-*-*-*",
        "screenInfo": "320-568-548-32-*-*-*",
        "capabilities": {
            "js": {
                "audio": True,
                "geolocation": True,
                "localStorage": "supported",
                "touch": True,
                "video": True,
                "webWorker": True
            },
            "css": {
                "textShadow": True,
                "textStroke": True,
                "boxShadow": True,
                "borderRadius": True,
                "borderImage": True,
                "opacity": True,
                "transform": True,
                "transition": True
            },
            "elapsed": 1
        },
        "referrer": "",
        "userAgent": user_agent,
        "location": oauth_url,
        "webDriver": None,
        "history": {
            "length": 1
        },
        "gpu": {
            "vendor": "Apple Inc.",
            "model": "Apple A9 GPU",
            "extensions": []
        },
        "math": {
            "tan": "-1.4214488238747243",
            "sin": "0.8178819121159085",
            "cos": "-0.5753861119575491"
        },
        "performance": {
            "timing": {
                "navigationStart": now_to_unix_ms(),
                "unloadEventStart": 0,
                "unloadEventEnd": 0,
                "redirectStart": 0,
                "redirectEnd": 0,
                "fetchStart": now_to_unix_ms(),
                "domainLookupStart": now_to_unix_ms(),
                "domainLookupEnd": now_to_unix_ms(),
                "connectStart": now_to_unix_ms(),
                "connectEnd": now_to_unix_ms(),
                "secureConnectionStart": now_to_unix_ms(),
                "requestStart": now_to_unix_ms(),
                "responseStart": now_to_unix_ms(),
                "responseEnd": now_to_unix_ms(),
                "domLoading": now_to_unix_ms(),
                "domInteractive": now_to_unix_ms(),
                "domContentLoadedEventStart": now_to_unix_ms(),
                "domContentLoadedEventEnd": now_to_unix_ms(),
                "domComplete": now_to_unix_ms(),
                "loadEventStart": now_to_unix_ms(),
                "loadEventEnd": now_to_unix_ms()
            }
        },
        "end": now_to_unix_ms(),
        "timeToSubmit": 108873,
        "form": {
            "email": {
                "keys": 0,
                "keyPressTimeIntervals": [],
                "copies": 0,
                "cuts": 0,
                "pastes": 0,
                "clicks": 0,
                "touches": 0,
                "mouseClickPositions": [],
                "keyCycles": [],
                "mouseCycles": [],
                "touchCycles": [],
                "width": 290,
                "height": 43,
                "checksum": "C860E86B",
                "time": 12773,
                "autocomplete": False,
                "prefilled": False
            },
            "password": {
                "keys": 0,
                "keyPressTimeIntervals": [],
                "copies": 0,
                "cuts": 0,
                "pastes": 0,
                "clicks": 0,
                "touches": 0,
                "mouseClickPositions": [],
                "keyCycles": [],
                "mouseCycles": [],
                "touchCycles": [],
                "width": 290,
                "height": 43,
                "time": 10353,
                "autocomplete": False,
                "prefilled": False
            }
        },
        "canvas": {
            "hash": -373378155,
            "emailHash": -1447130560,
            "histogramBins": []
        },
        "token": None,
        "errors": [],
        "metrics": [
            {
                "n": "fwcim-mercury-collector",
                "t": 0
            },
            {
                "n": "fwcim-instant-collector",
                "t": 0
            },
            {
                "n": "fwcim-element-telemetry-collector",
                "t": 2
            },
            {
                "n": "fwcim-script-version-collector",
                "t": 0
            },
            {
                "n": "fwcim-local-storage-identifier-collector",
                "t": 0
            },
            {
                "n": "fwcim-timezone-collector",
                "t": 0
            },
            {
                "n": "fwcim-script-collector",
                "t": 1
            },
            {
                "n": "fwcim-plugin-collector",
                "t": 0
            },
            {
                "n": "fwcim-capability-collector",
                "t": 1
            },
            {
                "n": "fwcim-browser-collector",
                "t": 0
            },
            {
                "n": "fwcim-history-collector",
                "t": 0
            },
            {
                "n": "fwcim-gpu-collector",
                "t": 1
            },
            {
                "n": "fwcim-battery-collector",
                "t": 0
            },
            {
                "n": "fwcim-dnt-collector",
                "t": 0
            },
            {
                "n": "fwcim-math-fingerprint-collector",
                "t": 0
            },
            {
                "n": "fwcim-performance-collector",
                "t": 0
            },
            {
                "n": "fwcim-timer-collector",
                "t": 0
            },
            {
                "n": "fwcim-time-to-submit-collector",
                "t": 0
            },
            {
                "n": "fwcim-form-input-telemetry-collector",
                "t": 4
            },
            {
                "n": "fwcim-canvas-collector",
                "t": 2
            },
            {
                "n": "fwcim-captcha-telemetry-collector",
                "t": 0
            },
            {
                "n": "fwcim-proof-of-work-collector",
                "t": 1
            },
            {
                "n": "fwcim-ubf-collector",
                "t": 0
            },
            {
                "n": "fwcim-timer-collector",
                "t": 0
            }
        ]
    }
    return json.dumps(meta_dict, separators=(',', ':'))



import logging
import pathlib
import re
from typing import Any, Optional, Union

from .localization import Locale


logger = logging.getLogger('audible.utils')


def test_convert(key: str, value: Any) -> Any:
    if key == "login_cookies":
        return _check_login_cookies(value) or value

    elif key == "adp_token":
        return _check_adp_token(value) or value

    elif key == "access_token":
        return _check_access_token(value) or value

    elif key == "refresh_token":
        return _check_refresh_token(value) or value

    elif key == "device_private_key":
        return _check_device_private_key(value) or value

    elif key == "expires":
        return _check_expires(value) or value

    elif key == "locale":
        return _check_locales(value) or value

    elif key == "filename":
        return _check_filename(value) or value

    elif key == "crypter":
        return _check_crypter(value) or value

    elif key == "encryption":
        return _check_encryption(value) or value

    else:
        return value


def _check_login_cookies(value) -> None:
    if not isinstance(value, dict):
        raise TypeError("type of login_cookies is not dict")
    if not set(map(type, value.values())) == {str}:
        raise TypeError("login_cookies have wrong format")


def _check_adp_token(value) -> None:
    if not isinstance(value, str):
        raise TypeError("type of adp_token is not str")

    parts = re.findall(r"\{.*?\}", value)
    if len(parts) != 5:
        raise ValueError("adp_token have wrong format")

    adp_token_as_dict = dict()
    for part in parts:
        search_pieces = re.search('\{(.*?)\:(.*?)\}', part)
        key = search_pieces.group(1)
        value = search_pieces.group(2)
        adp_token_as_dict[key] = value

        allowed_keys = {"enc", "key", "iv", "name", "serial"}
    if not adp_token_as_dict.keys() == allowed_keys:
        raise ValueError("adp_token have wrong format")


def _check_access_token(value) -> None:
    if not isinstance(value, str):
        raise TypeError("type of access_token is not str")

    if not value.startswith("Atna|"):
        raise ValueError('access_token have wrong format')


def _check_refresh_token(value) -> None:
    if not isinstance(value, str):
        raise TypeError("type of refresh_token is not str")
    if not value.startswith("Atnr|"):
        raise ValueError('refresh_token have wrong format')


def _check_device_private_key(value) -> None:
    if not isinstance(value, str):
        raise TypeError("type of device_private_key is not str")

    if not value.startswith("-----BEGIN RSA PRIVATE KEY-----"):
        raise ValueError('device_private_key have wrong format')

    if not value.endswith("-----END RSA PRIVATE KEY-----\n"):
        raise ValueError('device_private_key have wrong format')


def _check_expires(value) -> Optional[float]:
    if not isinstance(value, (int, float, str)):
        raise ValueError("type of device_private_key is not int or float")

    if isinstance(value, str):
        try:
            return float(value)
        except ValueError:
            raise ValueError("type of device_private_key is str "
                             "and can't be converted to float")


def _check_locales(value) -> Optional[Locale]:
    if isinstance(value, Locale):
        pass
    elif isinstance(value, str):
        return Locale(value.lower())
    else:
        raise TypeError("Locale error")


def _check_filename(value) -> Union[pathlib.Path, pathlib.WindowsPath]:
    if not isinstance(value, (pathlib.Path, pathlib.WindowsPath)):
        try:
            return pathlib.Path(value)
        except NotImplementedError:
            return pathlib.WindowsPath(value)
        except:
            raise Exception("File error")


def _check_crypter(value) -> None:
    # TODO: write checks
    pass


def _check_encryption(value) -> None:
    allowed_values = [False, "json", "bytes"]

    if not isinstance(value, (bool, str)):
        raise TypeError("encryption has wrong type")
    if value not in allowed_values:
        raise ValueError("encryption has wrong value")



import base64
from collections.abc import MutableMapping
from datetime import datetime, timedelta
import json
import logging
from typing import Any, Dict

import requests
import rsa

from .aescipher import AESCipher, detect_file_encryption
from .login import login
from .register import deregister as deregister_
from .register import register as register_
from .utils import test_convert


logger = logging.getLogger('audible.auth')


def refresh_access_token(refresh_token: str, domain: str) -> Dict[str, Any]:
    """
    Refresh access token with refresh token.
    Access tokens are valid for 60 mins.
    
    """
    
    body = {
        "app_name": "Audible",
        "app_version": "3.7",
        "source_token": refresh_token,
        "requested_token_type": "access_token",
        "source_token_type": "refresh_token"
    }

    resp = requests.post(
        f"https://api.amazon.{domain}/auth/token", data=body
    )
    resp.raise_for_status()
    resp_json = resp.json()

    expires_in_sec = int(resp_json["expires_in"])
    expires = (datetime.utcnow() + timedelta(seconds=expires_in_sec)).timestamp()

    return {
        "access_token": resp_json["access_token"],
        "expires": expires
    }


def user_profile(access_token: str, domain: str) -> Dict[str, Any]:
    """Returns user profile."""

    headers = {
        "Authorization": f"Bearer {access_token}"
    }

    resp = requests.get(
        f"https://api.amazon.{domain}/user/profile", headers=headers
    )
    resp.raise_for_status()

    return resp.json()


def sign_request(path: str, method: str, body: str, adp_token: str,
                 private_key: str) -> Dict[str, str]:
    """
    Helper function who creates a signed header for requests.

    :param path: the requested url path and query
    :param method: the request method (GET, POST, DELETE, ...)
    :param body: the http message body
    :param adp_token: the token is obtained after register as device
    :param private_key: the rsa key obtained after register as device
    """
    date = datetime.utcnow().isoformat("T") + "Z"

    data = f"{method}\n{path}\n{date}\n"
    if body:
        if isinstance(body, bytes):
            body = body.decode('utf-8')
        data += body
    data += "\n"
    data += adp_token

    key = rsa.PrivateKey.load_pkcs1(private_key)
    cipher = rsa.pkcs1.sign(data.encode(), key, "SHA-256")
    signed_encoded = base64.b64encode(cipher)

    signature = f"{signed_encoded.decode()}:{date}"

    return {
        "x-adp-token": adp_token,
        "x-adp-alg": "SHA256withRSA:1.0",
        "x-adp-signature": signature
    }


class BaseAuthenticator(MutableMapping):
    """Base Class for retrieve and handle credentials."""

    def __init__(self):
        raise NotImplementedError()

    def __getitem__(self, key):
        return self.__dict__[key]

    def __getattr__(self, attr):
        try:
            return self.__getitem__(attr)
        except KeyError:
            return None

    def __delitem__(self, key):
        del self.__dict__[key]

    def __setitem__(self, key, value):
        self.__dict__[key] = test_convert(key, value)

    def __setattr__(self, attr, value):
        self.__setitem__(attr, value)

    def __iter__(self):
        return iter(self.__dict__)

    def __len__(self):
        return len(self.__dict__)

    def __repr__(self):
        return f"{type(self).__name__}({self.__dict__})"

    def to_file(self, filename=None, password=None, encryption="default",
                indent=4, set_default=True, **kwargs):

        if filename:
            filename = test_convert("filename", filename)
        elif self.filename:
            filename = self.filename
        else:
            raise ValueError("No filename provided")

        if encryption != "default":
            encryption = test_convert("encryption", encryption)
        elif self.encryption:
            encryption = self.encryption
        else:
            raise ValueError("No encryption provided")
        
        body = {
            "login_cookies": self.login_cookies,
            "adp_token": self.adp_token,
            "access_token": self.access_token,
            "refresh_token": self.refresh_token,
            "device_private_key": self.device_private_key,
            "store_authentication_cookie": self.store_authentication_cookie,
            "device_info": self.device_info,
            "customer_info": self.customer_info,
            "expires": self.expires,
            "locale_code": self.locale.countryCode
        }
        json_body = json.dumps(body, indent=indent)

        if encryption is False:
            filename.write_text(json_body)
            crypter = None
        else:
            if password:
                crypter = test_convert("crypter", AESCipher(password, **kwargs))
            elif self.crypter:
                crypter = self.crypter
            else:
                raise ValueError("No password provided")

            crypter.to_file(json_body, filename=filename,
                            encryption=encryption, indent=indent)

        logger.info(f"saved data to file {filename}")

        if set_default:
            self.filename = filename
            self.encryption = encryption
            self.crypter = crypter

        logger.info(f"set filename {filename} as default")

    def re_login(self, username, password, captcha_callback=None,
                 otp_callback=None):

        login_device = login(
            username=username,
            password=password,
            countryCode = self.locale.countryCode,
            domain = self.locale.domain,
            marketPlaceId = self.locale.marketPlaceId,
            captcha_callback=captcha_callback,
            otp_callback=otp_callback
        )

        self.update(**login_device)

    def register_device(self):
        register_device = register_(access_token=self.access_token,
                                    domain=self.locale.domain)

        self.update(**register_device)

    def deregister_device(self, deregister_all: bool = False):
        return deregister_(access_token=self.access_token,
                           deregister_all=deregister_all,
                           domain=self.locale.domain)

    def refresh_access_token(self, force=False):
        if force or self.access_token_expired:
            refresh_data = refresh_access_token(
                refresh_token=self.refresh_token,
                domain=self.locale.domain
            )
    
            self.update(**refresh_data)
        else:
            print("Access Token not expired. No refresh nessessary. "
                  "To force refresh please use force=True")

    def refresh_or_register(self, force=False):
        try:
            self.refresh_access_token(force=force)
        except:
            try:
                self.deregister_device()
                self.register_device()
            except:
                raise Exception("Could not refresh client.")

    def user_profile(self):
        return user_profile(access_token=self.access_token,
                            domain=self.locale.domain)

    @property
    def access_token_expires(self):
        return datetime.fromtimestamp(self.expires) - datetime.utcnow()

    @property
    def access_token_expired(self):
        if datetime.fromtimestamp(self.expires) <= datetime.utcnow():
            return True
        else:
            return False


class LoginAuthenticator(BaseAuthenticator):
    """Authenticator class to retrieve credentials from login."""
    def __init__(self, username: str, password: str, locale, register=True,
                 captcha_callback=None, otp_callback=None):

        self.locale = locale

        resp = login(
            username=username,
            password=password,
            countryCode = self.locale.countryCode,
            domain = self.locale.domain,
            marketPlaceId = self.locale.marketPlaceId,
            captcha_callback=captcha_callback,
            otp_callback=otp_callback
        )
 
        logger.info(f"logged in to audible as {username}")

        if register:
            resp = register_(resp["access_token"], self.locale.domain)
        logger.info("registered audible device")

        self.update(**resp)


class FileAuthenticator(BaseAuthenticator):
    """Authenticator class to retrieve credentials from stored file."""
    def __init__(self, filename, password=None, locale=None,
                 encryption=None, **kwargs) -> None:

        self.filename = filename

        self.encryption = detect_file_encryption(self.filename)

        if self.encryption:
            self.crypter = AESCipher(password, **kwargs)
            file_data = self.crypter.from_file(self.filename, self.encryption)
        else:
            file_data = self.filename.read_text()

        json_data = json.loads(file_data)

        locale_code = json_data.pop("locale_code", None)
        locale = locale or locale_code
        self.locale = locale

        self.update(**json_data)

        logger.info((f"load data from file {self.filename} for "
                     f"locale {self.locale.countryCode}"))



from datetime import datetime, timedelta
from random import choices
import string
from typing import Any, Dict

import requests


def get_random_device_serial() -> str:
    """
    Generates a random text (str + int) with a length of 40 chars.
    
    Use of random serial prevents unregister device by other users
    with same `device_serial`.
    """
    return "".join(choices(string.ascii_uppercase + string.digits, k=40))


def register(access_token: str, domain: str) -> Dict[str, Any]:
    """
    Register a dummy audible device with access token and cookies
    from ``auth_login``.
    Returns important credentials needed for access audible api.
    """
    body = {
        "requested_token_type": [
            "bearer", "mac_dms", "website_cookies",
            "store_authentication_cookie"
        ],
        "cookies": {
            "website_cookies": [],
            "domain": f".amazon.{domain}"
        },
        "registration_data": {
            "domain": "Device",
            "app_version": "3.7",
            "device_serial": get_random_device_serial(),
            "device_type": "A2CZJZGLK2JJVM",
            "device_name": ("%FIRST_NAME%%FIRST_NAME_POSSESSIVE_STRING%%DUPE_"
                            "STRATEGY_1ST%Audible for iPhone"),
            "os_version": "12.3.1",
            "device_model": "iPhone",
            "app_name": "Audible"
        },
        "auth_data": {
            "access_token": access_token
        },
        "requested_extensions": ["device_info", "customer_info"]
    }

    resp = requests.post(
        f"https://api.amazon.{domain}/auth/register", json=body
    )

    resp_json = resp.json()
    if resp.status_code != 200:
        raise Exception(resp_json)

    success_response = resp_json["response"]["success"]

    tokens = success_response["tokens"]
    adp_token = tokens["mac_dms"]["adp_token"]
    device_private_key = tokens["mac_dms"]["device_private_key"]
    store_authentication_cookie = tokens["store_authentication_cookie"]
    access_token = tokens["bearer"]["access_token"]
    refresh_token = tokens["bearer"]["refresh_token"]
    expires_s = int(tokens["bearer"]["expires_in"])
    expires = (datetime.utcnow() + timedelta(seconds=expires_s)).timestamp()

    extensions = success_response["extensions"]
    device_info = extensions["device_info"]
    customer_info = extensions["customer_info"]

    login_cookies_new = dict()
    for cookie in tokens["website_cookies"]:
        login_cookies_new[cookie["Name"]] = cookie["Value"].replace(r'"', r'')

    return {
        "adp_token": adp_token,
        "device_private_key": device_private_key,
        "access_token": access_token,
        "refresh_token": refresh_token,
        "expires": expires,
        "login_cookies": login_cookies_new,
        "store_authentication_cookie": store_authentication_cookie,
        "device_info": device_info,
        "customer_info": customer_info
    }


def deregister(access_token: str, domain: str,
               deregister_all: bool = False) -> Dict[str, Any]:
    """
    Deregister device which was previously registered with `access_token`.

    `access_token` is valid until expiration. All other credentials will
    be invalid immediately.
    
    If `deregister_all=True` all registered devices will be deregistered.
    """
    body = {"deregister_all_existing_accounts": deregister_all}
    headers = {"Authorization": f"Bearer {access_token}"}

    resp = requests.post(
        f"https://api.amazon.{domain}/auth/deregister",
        json=body, headers=headers
    )

    resp_json = resp.json()
    if resp.status_code == 200:
        return resp_json
    else:
        raise Exception(resp_json)



import logging
from typing import Any, Dict, Optional
from urllib.parse import parse_qs, urlparse

from bs4 import BeautifulSoup
import requests


logger = logging.getLogger('audible.localization')


LOCALE_TEMPLATES = {
    "germany": {
        "countryCode": "de",
        "domain": "de",
        "marketPlaceId": "AN7V1F1VY261K"
    },
    "united_states": {
        "countryCode": "us",
        "domain": "com",
        "marketPlaceId": "AF2M0KC94RCEA"
    },
    "united_kingdom": {
        "countryCode": "uk",
        "domain": "co.uk",
        "marketPlaceId": "A2I9A3Q2GNFNGQ"
    },
    "france": {
        "countryCode": "fr",
        "domain": "fr",
        "marketPlaceId": "A2728XDNODOQ8T"
    },
    "canada": {
        "countryCode": "ca",
        "domain": "ca",
        "marketPlaceId": "A2CQZ5RBY40XE"
    }
}


def search_template(key: str, value: Any):
    for country in LOCALE_TEMPLATES:
        locale = LOCALE_TEMPLATES[country]
        if locale[key] == value:
            logger.debug(f"found locale for {country}")
            return locale

    logger.info(f"don\'t found {value} in {key}")
    return None


def autodetect_locale(domain: str) -> Dict[str, str]:
    """
    Try to automatically detect correct settings for marketplace.

    Needs the top level domain of the audible page to continue with
    (e.g. co.uk, co.jp) and returns results found.

    """
    domain = domain.lstrip(".")
    site = f"https://www.audible.{domain}"

    try:
        resp = requests.get(site)
    except requests.exceptions.ConnectionError:
        logger.warn(f"site {site} doesn\'t exists or Network Error occours")
        return None

    soup = BeautifulSoup(resp.text, "html.parser")

    login_link = soup.find("a", class_="ui-it-sign-in-link")["href"]
    parsed_link = urlparse(login_link)
    query_string = parse_qs(parsed_link.query)
    marketPlaceId = query_string['marketPlaceId'][0]
    countryCode = query_string['pageId'][0].split("_")[-1]

    return {
        "countryCode": countryCode,
        "domain": domain,
        "marketPlaceId": marketPlaceId
    }


class Locale:
    """
    Adjustments for the different marketplaces who are provided by audible.

    Look at `locales.json` for examples. You can try to
    ``autodetect_locale`` if your marketplace is not in `locales.json`.
    
    """
    def __init__(self, countryCode: Optional[str] = None,
                 domain: Optional[str] = None,
                 marketPlaceId: Optional[str] = None) -> None:

        if not all([countryCode, domain, marketPlaceId]):
            locale = None
            if countryCode:
                locale = search_template("countryCode", countryCode)
            elif domain:
                locale = search_template("domain", domain)

            if locale is None:
                raise Exception("can\'t find locale")

        self._countryCode = countryCode or locale["countryCode"]
        self._domain = domain or locale["domain"]
        self._marketPlaceId = marketPlaceId or locale["marketPlaceId"]

    def __repr__(self):
        return (f"Locale class for domain: {self.domain}, "
                f"marketplace: {self.marketPlaceId}")

    def to_dict(self) -> Dict[str, str]:
        return {
            "countryCode": self.countryCode,
            "domain": self.domain,
            "marketPlaceId": self.marketPlaceId
        }

    @property
    def countryCode(self) -> str:
        return self._countryCode

    @property
    def domain(self) -> str:
        return self._domain

    @property
    def marketPlaceId(self) -> str:
        return self._marketPlaceId



import asyncio
import json
import logging
from typing import Union, Optional
from urllib.parse import urlencode, urlparse


import aiohttp
import requests
import yarl

from .auth import sign_request, LoginAuthenticator, FileAuthenticator
from .errors import (BadRequest, NotFoundError, NotResponding, NetworkError,
                     ServerError, Unauthorized, UnexpectedError,
                     RatelimitError)


logger = logging.getLogger('audible.client')


class AudibleAPI:

    REQUEST_LOG = '{method} {url} has received {text}, has returned {status}'

    def __init__(self, auth: Optional[Union[LoginAuthenticator, FileAuthenticator]] = None,
                 session=None, is_async=False, **options) -> None:
        self.auth = auth
        self.adp_token = options.get("adp_token", auth.adp_token)
        self.device_private_key = options.get(
            "device_private_key", auth.device_private_key
        )

        self.is_async = is_async
        self.session = (session or (aiohttp.ClientSession() if is_async
                        else requests.Session()))
        self.headers = {
            "Accept": "application/json",
            "Content-Type": "application/json"
        }

        domain = options.get("domain", auth.locale.domain)
        self.api_root_url = options.get("url", f"https://api.audible.{domain}")

        self.timeout = options.get('timeout', 10)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.close()

    async def __aenter__(self):
        return self

    async def __aexit__(self, exc_type, exc_value, traceback):
        await self.session.close()

    def __repr__(self):
        return f"<AudibleAPI Client async={self.is_async}>"

    def close(self):
        return self.session.close()

    def _raise_for_status(self, resp, text, *, method=None):
        try:
            data = json.loads(text)
        except json.JSONDecodeError:
            data = text
        code = getattr(resp, 'status', None) or getattr(resp, 'status_code')

        logger.debug(self.REQUEST_LOG.format(
            method=method or resp.request_info.method, url=resp.url,
            text=text, status=code
        ))

        if 300 > code >= 200:  # Request was successful
            return data, resp  # value, response
        elif code == 400:
            raise BadRequest(resp, data)
        elif code in (401, 403):  # Unauthorized request - Invalid credentials
            raise Unauthorized(resp, data)
        elif code == 404:  # not found
            raise NotFoundError(resp, data)
        elif code == 429:
            raise RatelimitError(resp, data)
        elif code == 503:  # Maintainence
            raise ServerError(resp, data)
        else:
            raise UnexpectedError(resp, data)

    def _sign_request(self, method, url, params, json_data):
        path = urlparse(url).path
        query = urlencode(params)
        json_body = json.dumps(json_data) if json_data else None
    
        if query:
            path += f"?{query}"

        return sign_request(path, method, json_body, self.adp_token,
                            self.device_private_key)

    async def _arequest(self, method, url, **kwargs):
        params = kwargs.pop("params", {})
        json_data = kwargs.get('json', {})
        timeout = kwargs.pop('timeout', self.timeout)

        headers = kwargs.pop("headers", {})
        signed_headers = self._sign_request(
            method, url, params, json_data
        )
        headers.update(self.headers)
        headers.update(signed_headers)

        url += "?" + urlencode(params)
        url = yarl.URL(url, encoded=True)

        try:
            async with self.session.request(
                method, url, timeout=timeout, headers=headers, **kwargs
            ) as resp:
                return self._raise_for_status(resp, await resp.text())
        except asyncio.TimeoutError:
            raise NotResponding
        except aiohttp.ServerDisconnectedError:
            raise NetworkError

    def _request(self, method, url, **kwargs):
        if self.is_async:  # return a coroutine
            return self._arequest(method, url, **kwargs)
        params = kwargs.get("params", {})
        json_data = kwargs.get('json', {})
        timeout = kwargs.pop('timeout', self.timeout)

        headers = kwargs.pop("headers", {})
        signed_headers = self._sign_request(method, url, params, json_data)
        headers.update(self.headers)
        headers.update(signed_headers)

        try:
            with self.session.request(
                method, url, timeout=timeout, headers=headers, **kwargs
            ) as resp:
                return self._raise_for_status(resp, resp.text, method=method)
        except requests.Timeout:
            raise NotResponding
        except requests.ConnectionError:
            raise NetworkError

    def _split_kwargs(self, **kwargs):
        requests_kwargs = [
            "method", "url", "params", "data", "json", "headers", "cookies",
            "files", "auth", "timeout", "allow_redirects", "proxies", "verify",
            "stream", "cert"
        ]
        params = kwargs.pop("params", {})
        for key in list(kwargs.keys()):
            if key not in requests_kwargs:
                params[key] = kwargs.pop(key)

        return params, kwargs

    def get(self, path, api_version="1.0", **kwargs):
        params, kwargs = self._split_kwargs(**kwargs)
        url = "/".join((self.api_root_url, api_version, path))
        return self._request("GET", url, params=params, **kwargs)

    def post(self, path, body, api_version="1.0", **kwargs):
        params, kwargs = self._split_kwargs(**kwargs)
        url = "/".join((self.api_root_url, api_version, path))
        return self._request("POST", url, params=params, json=body, **kwargs)

    def delete(self, path, api_version="1.0", **kwargs):
        params, kwargs = self._split_kwargs(**kwargs)
        url = "/".join((self.api_root_url, api_version, path))
        return self._request("DELETE", url, params=params, **kwargs)



__title__ = 'audible'
__description__ = 'A(Sync) Interface for internal Audible API written in pure Python.'
__url__ = 'https://github.com/mkb79/audible'
__version__ = '0.2.1'
__author__ = 'mkb79'
__author_email__ = 'mkb79@hackitall.de'
__license__ = 'AGPL'
__status__ = 'Development'


__all__ = [
    "__title__", "__description__", "__url__", "__version__",
    "__author__", "__author_email__", "__license__", "__status__"
]



import logging
import pathlib
from typing import Union


LEVELS = {
    "debug": logging.DEBUG,
    "info": logging.INFO,
    "warn": logging.WARNING,
    "warning": logging.WARNING,
    "error": logging.ERROR,
    "critical": logging.CRITICAL
}


log_formatter = logging.Formatter(("%(asctime)s %(levelname)s [%(name)s] "
                                   "%(filename)s:%(lineno)d: %(message)s"))

logger = logging.getLogger("audible")
logger.setLevel(logging.CRITICAL)
logging.captureWarnings(True)


def set_global_level(level: Union[str, int]) -> None:
    """Set logging level for the main logger."""
    if isinstance(level, str):
        level = level.lower().strip()
        level = LEVELS.get(level, logging.NOTSET)
    elif not isinstance(level, int):
        raise TypeError((f"Level is from type {type(level)} but only "
                         f"str and int are allowed."))

    logger.setLevel(level)
    logger.log(logging.INFO, "set logging threshold to \"%s\"",
               logging.getLevelName(logger.level))


def set_console_logger(level: Union[str, int] = 0) -> None:
    """Set logging level for the stream handler."""
    stream_handler = logging.StreamHandler()
    stream_handler.setFormatter(log_formatter)
    stream_handler.set_name("ConsoleLogger")
    logger.addHandler(stream_handler)
    _setLevel(stream_handler, level)


def set_file_logger(filename: str, level: Union[str, int] = 0) -> None:
    """Set logging level and filename for the file handler."""
    try:
        filename = pathlib.Path(filename)
    except NotImplementedError:
        filename = pathlib.WindowsPath(filename)

    file_handler = logging.FileHandler(filename)
    file_handler.setFormatter(log_formatter)
    file_handler.set_name("FileLogger")
    logger.addHandler(file_handler)
    _setLevel(file_handler, level)


def _setLevel(handler, level: Union[str, int]) -> None:
    if isinstance(level, str):
        level = level.lower().strip()
        level = LEVELS.get(level, logging.NOTSET)
    elif not isinstance(level, int):
        raise TypeError((f"Level is from type {type(level)} but only "
                         f"str and int are allowed."))
        
    handler.setLevel(level)
    if handler.level < logger.level or logger.level == 0:
        logger.setLevel(handler.level)
    logger.log(logging.INFO, (f"set logging threshold for \"{handler.name}\" "
                              f"to \"{logging.getLevelName(handler.level)}\""))



from os import system
import pathlib
from setuptools import setup
import sys


# 'setup.py publish' shortcut.
if sys.argv[-1] == 'publish':
    system('python setup.py sdist bdist_wheel')
    system('twine upload dist/*')
    sys.exit()

if sys.version_info < (3, 6, 0):
    raise RuntimeError("audible requires Python 3.6.0+")

here = pathlib.Path(__file__).parent

packages = ['audible']

about = {}
exec((here / 'audible' / '_version.py').read_text('utf-8'), about)

long_description = (here / 'README.md').read_text('utf-8')

requires = (here / 'requirements.txt').read_text('utf-8').split()


setup(
    name=about['__title__'],
    version=about['__version__'],
    packages=packages,
    package_dir={'audible': 'audible'},
    description=about['__description__'],
    url=about['__url__'],
    license=about['__license__'],
    author=about['__author__'],
    author_email=about['__author_email__'],
    classifiers=[
         'Development Status :: 4 - Beta',
         'Intended Audience :: Developers',
         'License :: OSI Approved :: GNU Affero General Public License v3',
         'Programming Language :: Python :: 3.6',
         'Programming Language :: Python :: 3.7',
         'Programming Language :: Python :: 3.8'
    ],
    install_requires=requires,
    python_requires='>=3.6',
    keywords='Audible, API, async',
    include_package_data=True,
    long_description=long_description,
    long_description_content_type='text/markdown',
)
